/*
 * linux/include/asm-arm/proc-armv/mm-init.h
 *
 * Copyright (C) 1996 Russell King
 *
 * This contains the code to setup the memory map on an ARM v3 or v4 machine.
 * This is both processor & architecture specific, and requires some
 * more work to get it to fit into our separate processor and architecture
 * structure.
 */

/*
 * On vnc, we want the memory map set up so:
 *
 *   PHYS        VIRT
 * 00000000    00000000        Zero page
 * 000003ff    000003ff        Zero page end
 * 00000000    c0000000        Kernel and all physical memory
 * 03ffffff    c3ffffff        End of physical (64MB)
 * 88000000    d0000000        vga memory
 * e0000000    e0000000        IO start, including io, footbridge, flash, etc
 * 00000000    f0000000        Second non-cacheable mapping of physical mem
 * ffffffff    ffffffff        IO end
 *
 *
 * On ebsa, we want the memory map set up so:
 *
 *   PHYS	  VIRT
 * 00000000	00000000	Zero page
 * 000003ff	000003ff	Zero page end
 * 00000000	c0000000	Kernel and all physical memory
 * 01ffffff	c1ffffff	End of physical (32MB)
 * e0000000	e0000000	IO start
 * ffffffff	ffffffff	IO end
 *
 * On rpc, we want:
 *
 *   PHYS	  VIRT
 * 10000000	00000000	Zero page
 * 100003ff	000003ff	Zero page end
 * 10000000	c0000000	Kernel and all physical memory
 * 1fffffff	cfffffff	End of physical (32MB)
 * 02000000	d?000000	Screen memory (first image)
 * 02000000	d8000000	Screen memory (second image)
 * 00000000	df000000	StrongARM cache invalidation area
 * 03000000	e0000000	IO start
 * 03ffffff	e0ffffff	IO end
 *
 * We set it up using the section page table entries.
 */

#include <asm/arch/mmap.h>
#include <asm/pgtable.h>
 

#ifdef CONFIG_ARCH_VNC
extern unsigned long vnc_io_page_table[];
#define NR_IO_ENTRIES 32
#endif
#define V2P(x)	virt_to_phys(x)
#define PTE_SIZE (PTRS_PER_PTE * 4)

#define PMD_SECT	(PMD_TYPE_SECT | PMD_DOMAIN(DOMAIN_KERNEL) | PMD_SECT_CACHEABLE)

static inline void setup_swapper_dir (int index, unsigned long entry)
{
	pmd_t pmd;

	pmd_val(pmd) = entry;
	set_pmd (pmd_offset (swapper_pg_dir + index, 0), pmd);
}

static inline unsigned long setup_pagetables(unsigned long start_mem, unsigned long end_mem)
{
	unsigned long address;
	unsigned int spi, spi1;
	union { unsigned long l; unsigned long *p; } u;

	/* map in zero page */
	u.l = ((start_mem + (PTE_SIZE-1)) & ~(PTE_SIZE-1));
	start_mem = u.l + PTE_SIZE;
	memzero (u.p, PTE_SIZE);
	*u.p = V2P(PAGE_OFFSET) | PTE_CACHEABLE | PTE_TYPE_SMALL | PTE_AP_WRITE;
	setup_swapper_dir (0, V2P(u.l) | PMD_TYPE_TABLE | PMD_DOMAIN(DOMAIN_KERNEL));

	for (spi = 1; spi < (PAGE_OFFSET >> PGDIR_SHIFT); spi++)
		pgd_val(swapper_pg_dir[spi]) = 0;

	/* map in physical ram & kernel */
	address = PAGE_OFFSET;
	while (spi < end_mem >> PGDIR_SHIFT) {
		setup_swapper_dir (spi++,
				V2P(address) | PMD_SECT |
				PMD_SECT_BUFFERABLE | PMD_SECT_AP_WRITE);
		address += PGDIR_SIZE;
	}
	while (spi < PTRS_PER_PGD)
		pgd_val(swapper_pg_dir[spi++]) = 0;

		
#ifdef CONFIG_ARCH_VNC
		
	//wms: do a double-mapping of the physical memory at address 0xF0000000
	//for the benefit of hardware DMA using descriptor tables.
	//For now this should affect the IDE hard disk and Tulip Eth.100

	/* map again the physical ram as a non-cacheable area */
	spi1 = NONCACHED_PAGE_OFFSET >> PGDIR_SHIFT;
	spi = PAGE_OFFSET >> PGDIR_SHIFT;
	address = PAGE_OFFSET;
	while (spi < end_mem >> PGDIR_SHIFT) {
		setup_swapper_dir (spi1++,
				V2P(address) | PMD_TYPE_SECT | 
				PMD_DOMAIN(DOMAIN_KERNEL) | PMD_SECT_AP_WRITE);
		address += PGDIR_SIZE;
		spi++;
	}
//	while (spi < PTRS_PER_PGD)
//		pgd_val(swapper_pg_dir[spi1++]) = 0;
	//end of double mapping stuff
#endif		
		

	/*
	 * An area to invalidate the cache
	 */
	setup_swapper_dir (0xdf0, SAFE_ADDR | PMD_SECT | PMD_SECT_AP_READ);

	/* map in IO */
	address = IO_START;
	spi = IO_BASE >> PGDIR_SHIFT;
	pgd_val(swapper_pg_dir[spi-1]) = 0xc0000000 | PMD_TYPE_SECT |
			 PMD_DOMAIN(DOMAIN_KERNEL) | PMD_SECT_AP_WRITE;
#ifdef CONFIG_ARCH_VNC
	
        memcpy(&swapper_pg_dir[spi], vnc_io_page_table, sizeof(long) * NR_IO_ENTRIES);
//better replace calculated size with a struc sizeof, but does not compile...
//        memcpy(&swapper_pg_dir[spi], vnc_io_page_table, sizeof(vnc_io_page_table));
        for (spi = (0xD0000000 >> PGDIR_SHIFT), address = 0x88000000; spi < (0xD1000000 >> PGDIR_SHIFT); spi++, address += PGDIR_SIZE) {
                pgd_val(swapper_pg_dir[spi]) = address | PMD_TYPE_SECT | PMD_DOMAIN(DOMAIN_KERNEL) | PMD_SECT_AP_WRITE;
        }
/*
	//wms: Page required for the modem device. It decodes only the top nibble
	spi = 0xD1000000 >> PGDIR_SHIFT;
	pgd_val(swapper_pg_dir[spi]) = 0x90000000 | PMD_TYPE_SECT |
			 PMD_DOMAIN(DOMAIN_KERNEL) | PMD_SECT_AP_WRITE;
*/
/*
	//wms: Page required for the board rev reading by an ISA memory read
	spi = 0xD3000000 >> PGDIR_SHIFT;
	pgd_val(swapper_pg_dir[spi]) = 0x80000000 | PMD_TYPE_SECT |
			 PMD_DOMAIN(DOMAIN_KERNEL) | PMD_SECT_AP_WRITE;
*/

#else
	while (address < IO_START + IO_SIZE && address) {
		pgd_val(swapper_pg_dir[spi++]) = address |
						PMD_TYPE_SECT | PMD_DOMAIN(DOMAIN_KERNEL) |
						PMD_SECT_AP_WRITE;
		address += PGDIR_SIZE;
	}
#endif

#ifdef HAVE_MAP_VID_MEM
	map_screen_mem(0, 0, 0);
#endif

	flush_cache_all();
	return start_mem;
}

static inline void mark_usable_memory_areas(unsigned long *start_mem, unsigned long end_mem)
{
	unsigned long smem;

	*start_mem = smem = PAGE_ALIGN(*start_mem);

	while (smem < end_mem) {
	    	clear_bit(PG_reserved, &mem_map[MAP_NR(smem)].flags);
    		smem += PAGE_SIZE;
	}
}	

